import operator
from functools import reduce

from access_mopper.derivations.calc_atmos import level_to_height
from access_mopper.derivations.calc_land import (
    average_tile,
    calc_landcover,
    calc_topsoil,
    extract_tilefrac,
)

custom_functions = {
    "add": lambda *args: reduce(operator.add, args),
    "subtract": lambda a, b: a - b,
    "multiply": lambda a, b: a * b,
    "divide": lambda a, b: a / b,
    "power": lambda a, b: a**b,
    "sum": lambda x, **kwargs: x.sum(**kwargs),
    "mean": lambda *args: sum(args) / len(args),
    "kelvin_to_celsius": lambda x: x - 273.15,
    "celsius_to_kelvin": lambda x: x + 273.15,
    "level_to_height": level_to_height,
    "calc_topsoil": calc_topsoil,
    "calc_landcover": calc_landcover,
    "extract_tilefrac": extract_tilefrac,
    "average_tile": average_tile,
}


def evaluate_expression(expr, context):
    if isinstance(expr, dict):
        if "literal" in expr:
            return expr["literal"]
        op = expr["operation"]
        args = [
            evaluate_expression(arg, context)
            for arg in expr.get("args", expr.get("operands", []))
        ]
        kwargs = {
            k: evaluate_expression(v, context)
            for k, v in expr.get("kwargs", {}).items()
        }
        return custom_functions[op](*args, **kwargs)

    elif isinstance(expr, list):
        # Recursively evaluate items in the list
        return [evaluate_expression(item, context) for item in expr]

    elif isinstance(expr, str):
        # Lookup variable name in context
        return context[expr]

    elif isinstance(expr, (int, float)):
        return expr

    else:
        raise ValueError(f"Unsupported expression: {expr}")



"""
Smoke tests for ACCESS-MOPPeR.

This module contains basic smoke tests to ensure the main components
can be imported and initialized correctly. These are quick tests that
verify basic functionality without requiring extensive test data.
"""

import importlib.resources as resources
import tempfile
from pathlib import Path

import pytest

# Try to import optional dependencies gracefully
try:
    import pandas as pd

    PANDAS_AVAILABLE = True
except ImportError:
    PANDAS_AVAILABLE = False

try:
    from access_mopper import ACCESS_ESM_CMORiser

    ACCESS_MOPPER_AVAILABLE = True
except ImportError:
    ACCESS_MOPPER_AVAILABLE = False


DATA_DIR = Path(__file__).parent / "data"


@pytest.mark.skipif(not ACCESS_MOPPER_AVAILABLE, reason="ACCESS-MOPPeR not available")
def test_import_access_mopper():
    """Test that ACCESS_ESM_CMORiser can be imported."""
    assert ACCESS_ESM_CMORiser is not None


def test_test_data_exists():
    """Test that essential test data files exist."""
    test_file = DATA_DIR / "esm1-6/atmosphere/aiihca.pa-101909_mon.nc"
    # Use skipif for optional test data rather than failing
    if not test_file.exists():
        pytest.skip("Test data file not available")
    assert test_file.exists()


@pytest.mark.skipif(not PANDAS_AVAILABLE, reason="pandas not available")
def test_mapping_files_accessible():
    """Test that CMOR mapping files can be accessed."""
    mapping_files = [
        "Mappings_CMIP6_Amon.json",
        "Mappings_CMIP6_Lmon.json",
        "Mappings_CMIP6_Emon.json",
    ]

    for mapping_file in mapping_files:
        try:
            with (
                resources.files("access_mopper.mappings")
                .joinpath(mapping_file)
                .open() as f
            ):
                data = pd.read_json(f, orient="index")
                assert not data.empty, f"Empty mapping file: {mapping_file}"
        except Exception as e:
            pytest.fail(f"Cannot access mapping file {mapping_file}: {e}")


@pytest.mark.skipif(not ACCESS_MOPPER_AVAILABLE, reason="ACCESS-MOPPeR not available")
def test_cmoriser_initialization():
    """Test basic CMORiser initialization with minimal parameters."""
    try:
        # Use secure temporary directory instead of hardcoded /tmp
        with tempfile.TemporaryDirectory() as temp_dir:
            cmoriser = ACCESS_ESM_CMORiser(
                input_paths=["dummy.nc"],  # File doesn't need to exist for init test
                compound_name="Amon.tas",
                experiment_id="historical",
                source_id="ACCESS-ESM1-5",
                variant_label="r1i1p1f1",
                grid_label="gn",
                activity_id="CMIP",
                output_path=temp_dir,
            )

            # Test that basic attributes are set correctly
            assert cmoriser.experiment_id == "historical"
            assert cmoriser.compound_name == "Amon.tas"
            assert cmoriser.source_id == "ACCESS-ESM1-5"

    except Exception as e:
        pytest.fail(f"CMORiser initialization failed: {e}")


@pytest.mark.skipif(
    not ACCESS_MOPPER_AVAILABLE
    or not (
        Path(__file__).parent / "data/esm1-6/atmosphere/aiihca.pa-101909_mon.nc"
    ).exists(),
    reason="ACCESS-MOPPeR or test data not available",
)
def test_basic_cmorisation_workflow():
    """Test basic CMORisation workflow with a simple variable.

    This is a lightweight smoke test for the full workflow.
    More comprehensive tests are in the integration test modules.
    """
    test_file = DATA_DIR / "esm1-6/atmosphere/aiihca.pa-101909_mon.nc"

    # Use a simple, commonly available variable for smoke test
    parent_config = {
        "parent_experiment_id": "piControl",
        "parent_activity_id": "CMIP",
        "parent_source_id": "ACCESS-ESM1-5",
        "parent_variant_label": "r1i1p1f1",
        "parent_time_units": "days since 0001-01-01 00:00:00",
        "parent_mip_era": "CMIP6",
        "branch_time_in_child": 0.0,
        "branch_time_in_parent": 54786.0,
        "branch_method": "standard",
    }

    try:
        # Use secure temporary directory instead of hardcoded /tmp
        with tempfile.TemporaryDirectory() as temp_dir:
            cmoriser = ACCESS_ESM_CMORiser(
                input_paths=test_file,
                compound_name="Amon.tas",  # Use tas as it's commonly available
                experiment_id="historical",
                source_id="ACCESS-ESM1-5",
                variant_label="r1i1p1f1",
                grid_label="gn",
                activity_id="CMIP",
                parent_info=parent_config,
                output_path=temp_dir,
            )

            # Just test that run() doesn't crash - don't check output quality here
            cmoriser.run()

            # Basic check that some processing occurred
            assert hasattr(
                cmoriser, "cmor_ds"
            ), "CMORiser should have cmor_ds attribute after run()"

    except Exception as e:
        # For smoke tests, we want to know what failed but not necessarily fail the test
        pytest.skip(f"Smoke test skipped due to: {e}")



#!/usr/bin/env python
# Copyright 2024 ARC Centre of Excellence for Climate Extremes
# author: Paola Petrelli <paola.petrelli@utas.edu.au>
# author: Sam Green <sam.green@unsw.edu.au>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This is the ACCESS Model Output Post Processor, derived from the APP4
# originally written for CMIP5 by Peter Uhe and dapted for CMIP6 by Chloe Mackallah
# ( https://doi.org/10.5281/zenodo.7703469 )
#
# last updated 10/10/2024
#
# This file contains a collection of utilities to help calculate derived variables
# from ACCESS model output.
# Initial functions' definitions were based on APP4 modified to work with Xarray.


import json
import logging

import click
import numpy as np
import xarray as xr
from mopdb.utils import MopException

# Global Variables
# ----------------------------------------------------------------------

ice_density = 900  # kg/m3
snow_density = 300  # kg/m3

rd = 287.1
cp = 1003.5
p_0 = 100000.0
g_0 = 9.8067  # gravity constant
R_e = 6.378e06
# ----------------------------------------------------------------------


@click.pass_context
def time_resample(ctx, var, rfrq, tdim, sample="down", stats="mean"):
    """
    Resamples the input variable to the specified frequency using
    specified statistic.

    Resample is used with the options:
    origin = 'start_day'
    closed = 'right'
    This puts the time label to the start of the interval and
    offset is applied to get a centered time label.
    The `rfrq` valid lables are described here:
    https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#period-aliases

    Parameters
    ----------
    ctx : click context
        Includes obj dict with 'cmor' settings, exp attributes
    var : xarray.DataArray
        Variable to resample.
    rfrq : str
        Resample frequency see above for valid inputs.
    tdim: str
        The name of the time dimension
    sample : str
        The type of resampling to perform. Valid inputs are 'up' for
        upsampling or 'down' for downsampling. (default down)
    stats : str
        The reducing function to follow resample: mean, min, max, sum.
        (default mean)

    Returns
    -------
    vout : xarray.DataArray or xarray.Dataset
        The resampled variable.

    Raises
    ------
    ValueError
        If the input variable is not a valid Xarray object.
    ValueError
        If the sample parameter is not 'up' or 'down'.

    """
    var_log = logging.getLogger(ctx.obj["var_log"])
    if not isinstance(var, xr.DataArray):
        raise MopException("'var' must be a valid Xarray DataArray")
    valid_stats = ["mean", "min", "max", "sum"]
    if stats not in valid_stats:
        var_log.error(f"Resample unrecognised stats {stats}")
        raise MopException(f"{stats} not in valid list: {valid_stats}.")
    offset = {
        "30m": [15, "min"],
        "h": [30, "min"],
        "3h": [90, "min"],
        "6h": [3, "h"],
        "12h": [6, "h"],
        "D": [12, "h"],
        "7D": [84, "h"],
        "10D": [5, "D"],
        "M": [15, "D"],
        "Y": [6, "M"],
        "10Y": [5, "Y"],
    }
    if sample == "down":
        try:
            vout = var.resample({tdim: rfrq}, origin="start_day", closed="right")
            method = getattr(vout, stats)
            vout = method()
            half, tunit = offset[rfrq][:]
            vout = vout.assign_coords(
                {tdim: xr.CFTimeIndex(vout[tdim].values).shift(half, tunit)}
            )
        except Exception as e:
            var_log.error(f"Resample error: {e}")
            raise MopException(f"{e}")
    elif sample == "up":
        try:
            vout = var.resample({tdim: rfrq}).interpolate("linear")
        except Exception as e:
            var_log.error(f"Resample error: {e}")
            raise MopException(f"{e}")
    else:
        var_log.error("Resample can only be up or down")
        raise MopException("Sample is expected to be up or down")
    return vout


def add_axis(var, name, value):
    """Returns the same variable with an extra singleton axis added

    Parameters
    ----------
    var : Xarray DataArray
        Variable to modify
    name : str
        cmor name for axis
    value : float
        value of the new singleton dimension

    Returns
    -------
    var : Xarray DataArray
        Same variable with added axis at start

    """
    var = var.expand_dims(dim={name: float(value)})
    return var


@click.pass_context
def sum_vars(ctx, varlist):
    """Returns sum of all variables in list
    Parameters
    ----------
    varlist : list(xarray.DataArray)
        Variables to sum

    Returns
    -------
    varout : xarray.DataArray
        Sum of input variables

    """
    # first check that dimensions are same for all variables
    varout = varlist[0]
    for v in varlist[1:]:
        varout = varout + v
    return varout


@click.pass_context
def rename_coord(ctx, var1, var2, ndim, override=False):
    """If coordinates in ndim position are different, renames var2
    coordinates as var1.

    ctx : click context
        Includes obj dict with 'cmor' settings, exp attributes

    :meta private:
    """
    var_log = logging.getLogger(ctx.obj["var_log"])
    coord1 = var1.dims[ndim]
    coord2 = var2.dims[ndim]
    if coord1 != coord2:
        var_log.debug(f"{var1.name}, {var2.name}: {coord1}, {coord2}")
        var2 = var2.rename({coord2: coord1})
        if "bounds" in var1[coord1].attrs.keys():
            var2[coord1].attrs["bounds"] = var1[coord1].attrs["bounds"]
        override = True
    return var2, override


@click.pass_context
def get_ancil_var(ctx, ancil, varname):
    """Opens the ancillary file and get varname

    ctx : click context
        Includes obj dict with 'cmor' settings, exp attributes

    Returns
    -------
    ctx : click context obj
        Dictionary including 'cmor' settings and attributes for experiment
        Automatically passed
    var : Xarray DataArray
        selected variable from ancil file

    :meta private:
    """
    f = xr.open_dataset(f"{ctx.obj['ancil_path']}/" + f"{ctx.obj[ancil]}")
    var = f[varname]

    return var


@click.pass_context
def get_plev(ctx, levnum):
    """Read pressure levels from .._coordinate.json file

    ctx : click context
        Includes obj dict with 'cmor' settings, exp attributes

    Returns
    -------
    ctx : click context obj
        Dictionary including 'cmor' settings and attributes for experiment
        Automatically passed
    levnum : str
        Indicates pressure levels to load, corresponds to plev#levnum axis

    :meta private:
    """
    fpath = f"{ctx.obj['tpath']}/{ctx.obj['_AXIS_ENTRY_FILE']}"
    with open(fpath, "r") as jfile:
        data = json.load(jfile)
    axis_dict = data["axis_entry"]
    plev = np.array(axis_dict[f"plev{levnum}"]["requested"])
    plev = plev.astype(float)
    return plev



import os
import tempfile

import numpy as np
import requests
import xarray as xr
from tqdm import tqdm


class Supergrid:
    def __init__(self, nominal_resolution: str):
        """Initialize the Supergrid class with a specified nominal resolution."""

        self.nominal_resolution = nominal_resolution
        self.supergrid_path = self.get_supergrid_path(nominal_resolution)
        self.load_supergrid(self.supergrid_path)

    def get_supergrid_path(self, nominal_resolution: str) -> str:
        """Get the path to the supergrid file based on the nominal resolution.
        If the file is not found on Gadi, it will attempt to download it from Google Drive.
        """
        if not self.nominal_resolution:
            raise ValueError("nominal_resolution must be provided")
        # Mapping nominal resolution to file names
        supergrid_filenames = {
            "100 km": "mom1deg.nc",
            "25 km": "mom025deg.nc",
            "10 km": "mom01deg.nc",
        }

        if nominal_resolution not in supergrid_filenames:
            raise ValueError(
                f"Unknown or unsupported nominal resolution: {nominal_resolution}"
            )

        supergrid_filename = supergrid_filenames[nominal_resolution]
        gadi_supergrid_dir = "/g/data/xp65/public/apps/access_mopper_data/grids"
        gadi_supergrid_path = os.path.join(gadi_supergrid_dir, supergrid_filename)

        # Check if running on Gadi and file exists
        if os.path.exists(gadi_supergrid_path):
            supergrid_path = gadi_supergrid_path
        else:
            # Not on Gadi or file not available, download from Google Drive
            # Mapping nominal resolution to Google Drive file IDs
            gdrive_file_ids = {
                "100 km": "1Ito5EspxaICiTD1cfzcpcWTGNYg29fQf",
                "25 km": "1aNO1Y7HeU4YHjPi1Wsw_xRbp-SQG3NoA",
                "10 km": "GOOGLE_DRIVE_FILE_ID_FOR_10KM",
            }
            file_id = gdrive_file_ids[nominal_resolution]
            tmp_dir = tempfile.gettempdir()
            supergrid_path = os.path.join(tmp_dir, supergrid_filename)
            if not os.path.exists(supergrid_path):
                try:

                    def download_from_gdrive(file_id, dest_path):
                        # Download files from Google Drive (no token handling)
                        URL = (
                            f"https://drive.google.com/uc?export=download&id={file_id}"
                        )
                        with requests.get(URL, stream=True) as response:
                            response.raise_for_status()
                            total = int(response.headers.get("content-length", 0))
                            with (
                                open(dest_path, "wb") as f,
                                tqdm(
                                    total=total,
                                    unit="B",
                                    unit_scale=True,
                                    desc=f"Downloading {os.path.basename(dest_path)}",
                                ) as pbar,
                            ):
                                for chunk in response.iter_content(32768):
                                    if chunk:
                                        f.write(chunk)
                                        pbar.update(len(chunk))

                    download_from_gdrive(file_id, supergrid_path)
                except Exception as e:
                    raise RuntimeError(
                        f"Could not download supergrid file for {nominal_resolution}: {e}"
                    )
        return supergrid_path

    def load_supergrid(self, supergrid_file: str):
        """Load the supergrid dataset from the specified file."""
        if not supergrid_file:
            raise ValueError("supergrid_file must be provided")

        self.supergrid = xr.open_dataset(supergrid_file).rename_dims(
            {"nxp": "i_full", "nyp": "j_full"}
        )
        self.supergrid = self.supergrid.rename_vars({"x": "x_full", "y": "y_full"})
        self.xt = self.supergrid["x_full"][1::2, 1::2]
        self.yt = self.supergrid["y_full"][1::2, 1::2]
        self.xu = self.supergrid["x_full"][1::2, ::2]
        self.yu = self.supergrid["y_full"][1::2, ::2]
        self.xv = self.supergrid["x_full"][::2, 1::2]
        self.yv = self.supergrid["y_full"][::2, 1::2]
        self.xq = self.supergrid["x_full"][::2, ::2]
        self.yq = self.supergrid["y_full"][::2, ::2]

    def extract_grid(self, grid_type: str):
        if grid_type == "T":
            x = self.xt
            y = self.yt
            corners_x = self.xq
            corners_y = self.yq
        elif grid_type == "U":
            x = self.xu
            y = self.yu
            corners_x = self.supergrid["x_full"]
            corners_y = self.supergrid["y_full"]
        elif grid_type == "V":
            x = self.xv
            y = self.yv
            corners_x = self.supergrid["x_full"]
            corners_y = self.supergrid["y_full"]
        elif grid_type == "Q":
            x = self.xq
            y = self.yq
            corners_x = self.xq
            corners_y = self.yq
        else:
            raise ValueError(f"Unsupported grid_type: {grid_type}")

        corners_x = (corners_x + 360) % 360

        i_coord = xr.DataArray(
            np.arange(x.shape[1]),
            dims="i",
            name="i",
            attrs={"long_name": "cell index along first dimension", "units": "1"},
        )
        j_coord = xr.DataArray(
            np.arange(y.shape[0]),
            dims="j",
            name="j",
            attrs={"long_name": "cell index along second dimension", "units": "1"},
        )
        vertices = xr.DataArray(np.arange(4), dims="vertices", name="vertices")

        lat = xr.DataArray(y, dims=("j", "i"), name="latitude")
        lon = xr.DataArray((x + 360) % 360, dims=("j", "i"), name="longitude")

        lat_bnds = (
            xr.concat(
                [
                    corners_y[:-1, :-1].expand_dims(vertices=[0]),
                    corners_y[:-1, 1:].expand_dims(vertices=[1]),
                    corners_y[1:, 1:].expand_dims(vertices=[2]),
                    corners_y[1:, :-1].expand_dims(vertices=[3]),
                ],
                dim="vertices",
            )
            .rename({"j_full": "j", "i_full": "i"})
            .transpose("j", "i", "vertices")
            .rename("vertices_latitude")
        )

        lon_bnds = (
            xr.concat(
                [
                    corners_x[:-1, :-1].expand_dims(vertices=[0]),
                    corners_x[:-1, 1:].expand_dims(vertices=[1]),
                    corners_x[1:, 1:].expand_dims(vertices=[2]),
                    corners_x[1:, :-1].expand_dims(vertices=[3]),
                ],
                dim="vertices",
            )
            .rename({"j_full": "j", "i_full": "i"})
            .transpose("j", "i", "vertices")
            .rename("vertices_longitude")
        )

        return {
            "i": i_coord,
            "j": j_coord,
            "vertices": vertices,
            "latitude": lat,
            "longitude": lon,
            "vertices_latitude": lat_bnds,
            "vertices_longitude": lon_bnds,
        }
